import logging

from monopoly.config import (
    CreditStatementConfig,
    DebitStatementConfig,
    PdfConfig,
    settings,
)
from monopoly.constants import (
    BankNames,
    CreditTransactionPatterns,
    DebitTransactionPatterns,
    EncryptionIdentifier,
    MetadataIdentifier,
    StatementBalancePatterns,
)

from ..base import ProcessorBase

logger = logging.getLogger(__name__)


class Ocbc(ProcessorBase):
    credit_config = CreditStatementConfig(
        bank_name=BankNames.OCBC,
        statement_date_pattern=r"(\d{2}\-\d{2}\-\d{4})",
        prev_balance_pattern=StatementBalancePatterns.OCBC,
        transaction_pattern=CreditTransactionPatterns.OCBC,
    )

    debit_config = DebitStatementConfig(
        bank_name=BankNames.OCBC,
        statement_date_pattern=r"(\d+\s[A-Za-z]{3}\s\d{4})",
        debit_statement_identifier=r"(Withdrawal.*Deposit.*Balance)",
        transaction_pattern=DebitTransactionPatterns.OCBC,
        multiline_transactions=True,
    )

    pdf_config = PdfConfig(
        passwords=settings.ocbc_pdf_passwords,
    )

    identifiers = [
        EncryptionIdentifier(
            pdf_version=1.4, algorithm=4, revision=4, length=128, permissions=-1036
        ),
        MetadataIdentifier(
            creator="pdfgen", producer="Streamline PDFGen for OCBC Group"
        ),
    ]



import logging
import re
from functools import cached_property
from typing import Any

import fitz

from monopoly.config import CreditStatementConfig
from monopoly.constants import AccountType, StatementFields
from monopoly.pdf import PdfPage

from .base import BaseStatement, SafetyCheckError, Transaction

logger = logging.getLogger(__name__)


class CreditStatement(BaseStatement):
    """
    A dataclass representation of a credit statement
    """

    def __init__(
        self,
        document: fitz.Document,
        pages: list[PdfPage],
        credit_config: CreditStatementConfig,
    ):
        self.config = credit_config
        self.statement_type = AccountType.CREDIT
        super().__init__(document, pages, credit_config)

    @cached_property
    def prev_month_balance(self) -> list[Transaction]:
        """
        Returns the previous month's statement balance as a transaction,
        if it exists in the statement.

        The date is later replaced with a more accurate date by the statement processor.
        """
        prev_balances = []
        if prev_balance_pattern := self.config.prev_balance_pattern:
            raw_text = "".join(page.raw_text for page in self.pages)
            matches = re.finditer(prev_balance_pattern, raw_text)

            for match in matches:
                groupdict: dict[str, Any] = match.groupdict()
                groupdict[StatementFields.TRANSACTION_DATE] = "1900-01-01"
                prev_balances.append(Transaction(**groupdict))

        return prev_balances

    @cached_property
    def account_type(self) -> str:
        return AccountType.CREDIT

    def _inject_prev_month_balance(self, transactions: list[Transaction]):
        """
        Injects the previous month's balance as a transaction, if it exists
        """
        if self.prev_month_balance:
            first_transaction_date = transactions[0].transaction_date
            for prev_month_balance in self.prev_month_balance:
                prev_month_balance.transaction_date = first_transaction_date
                transactions.insert(0, prev_month_balance)
        return transactions

    @cached_property
    def transactions(self) -> list[Transaction]:
        transactions = self._inject_prev_month_balance(super().transactions)
        return transactions

    def perform_safety_check(self) -> bool:
        """Checks that the total sum of all transactions is present
        somewhere within the document

        Text is re-extracted from the page, as some bank-specific bounding-box
        configurations (e.g. HSBC) may preclude the total from being extracted

        Returns `False` if the total does not exist in the document.
        """
        df = self.df
        amount = StatementFields.AMOUNT
        numbers = self.get_all_numbers_from_document()

        total_amount = abs(round(df[amount].sum(), 2))

        result = total_amount in numbers
        if not result:
            raise SafetyCheckError(
                f"Total amount {total_amount} cannot be found in credit statement"
            )

        return result



import traceback
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import Collection, Iterable, Optional

import click
from pydantic.dataclasses import dataclass
from tqdm import tqdm

from monopoly.processors import detect_processor


@dataclass
class Result:
    """
    Stores information about processed bank statement
    """

    source_file_name: str
    target_file_name: Optional[str] = None
    error_info: Optional[dict[str, str]] = None


@dataclass
class Report:
    """
    A helper class for parsing and displaying processed bank statements.
    """

    results: list[Result]

    @property
    def processed_results(self) -> list[Result]:
        return sorted(
            [r for r in self.results if not r.error_info],
            key=lambda x: x.source_file_name,
        )

    @property
    def errored_results(self) -> list[Result]:
        return sorted(
            [r for r in self.results if r.error_info], key=lambda x: x.source_file_name
        )

    @property
    def number_processed(self) -> int:
        return len(self.processed_results)

    @property
    def number_errored(self) -> int:
        return len(self.errored_results)

    def display_report(self) -> None:
        """
        Parses all results, displaying the number of successfully
        processed statements and any errors.
        """
        if self.number_errored > 0:
            error_msg = (
                f"{self.number_errored} statement(s) had errors while processing"
            )
            click.echo(click.style(error_msg, fg="red", bold=True))
        if self.number_processed > 0:
            changed_msg = f"{self.number_processed} statement(s) processed"
            click.echo(click.style(changed_msg, bold=True))

        for res in self.errored_results:
            click.echo(
                click.style(
                    f"{res.source_file_name} -- "
                    f"{res.error_info['message'][:120]}",  # type: ignore
                    fg="red",
                )
            )

        for res in self.processed_results:
            click.echo(f"{res.source_file_name} -> {res.target_file_name}")


def process_statement(
    file: Path, output_directory: Optional[Path], print_df: bool
) -> Optional[Result]:
    """
    Extracts, transforms, and loads transactions from bank statements.

    Parameters:
        file: The path to the bank statement file.
        output_directory: The directory to save the processed statement.
            Defaults to the parent directory of the input file if not provided.
        print_df: If True, the transformed DataFrame is printed to the console
            in a tabular format. No file is generated in this case.

    Returns:
        Optional[Result]: If print_df is False, returns a Result object containing
        information about the processed statement. If an error occurs during processing,
        returns a Result object with error information.
    """

    try:
        processor = detect_processor(file)
        statement = processor.extract()
        transformed_df = processor.transform(statement)

        if print_df:
            click.echo(f"{file.name}")
            click.echo(transformed_df.to_markdown(tablefmt="psql", numalign="right"))
            click.echo()
            return None

        # saves processed statements to the same directory by default
        if not output_directory:
            output_directory = file.parent

        output_file = processor.load(transformed_df, statement, output_directory)
        return Result(file.name, output_file.name)

    except Exception as err:  # pylint: disable=broad-exception-caught
        error_info = {
            "message": f"{type(err).__name__}: {str(err)}",
            "traceback": traceback.format_exc(),
        }
        return Result(file.name, error_info=error_info)


def run(
    input_files: Collection[Path],
    output_directory: Optional[Path] = None,
    print_df: bool = False,
):
    """
    Process a collection of input files concurrently

    If any statements are processed successfully or encounter errors, a Report object
    is created and its display_report() method is called to provide a summary
    of the processing outcomes.
    """
    with ProcessPoolExecutor() as executor:
        results = list(
            tqdm(
                executor.map(
                    process_statement,
                    input_files,
                    [output_directory] * len(input_files),
                    [print_df] * len(input_files),
                ),
                total=len(input_files),
                desc="Processing statements",
                leave=False,
                delay=0.5,
                ncols=80,
                bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt}",
            )
        )

    if any(results):
        # filter out null values, for cases where print_df is True
        # and processing errors occur to avoid pydantic validation errors
        report = Report([res for res in results if res])
        report.display_report()


def get_statement_paths(files: Iterable[Path]) -> set[Path]:
    """
    Recursively collects paths to PDF files from a given collection of paths.
    """
    matched_files = set()
    for path in files:
        if path.is_file() and str(path).endswith(".pdf"):
            matched_files.add(path)

        if path.is_dir():
            matched_files |= get_statement_paths(path.iterdir())

    return matched_files


@click.command()
@click.version_option(package_name="monopoly-sg")
@click.argument(
    "files",
    nargs=-1,
    type=click.Path(exists=True, allow_dash=True, resolve_path=True, path_type=Path),
)
@click.option(
    "-o",
    "--output",
    type=click.Path(exists=True, allow_dash=True, resolve_path=True, path_type=Path),
    help="Specify output folder",
)
@click.option(
    "-p",
    "--pprint",
    is_flag=True,
    help="Print the processed statement(s)",
)
@click.pass_context
def monopoly(ctx: click.Context, files: list[Path], output: Path, pprint: bool):
    """
    Monopoly helps convert your bank statements from PDF to CSV.

    A file or directory can be passed in via the FILES argument
    """
    if files:
        matched_files = get_statement_paths(files)

        if matched_files:
            run(matched_files, output, pprint)
            ctx.exit(0)

        else:
            click.echo(
                click.style("Could not find .pdf files", fg="yellow", bold=True),
                err=True,
            )
            ctx.exit(1)
    else:
        show_welcome_message()


def show_welcome_message():
    art = r"""
     __  __                               _
    |  \/  | ___  _ __   ___  _ __   ___ | |_   _
    | |\/| |/ _ \| '_ \ / _ \| '_ \ / _ \| | | | |
    | |  | | (_) | | | | (_) | |_) | (_) | | |_| |
    |_|  |_|\___/|_| |_|\___/| .__/ \___/|_|\__, |
                             |_|            |___/
    """

    click.echo(art)
    message = "Monopoly helps convert your bank statements from PDF to CSV"
    click.echo(message)
    commands = [
        (
            "monopoly .",
            "process all statements nested in current dir",
        ),
        (
            "monopoly path/to/file.pdf",
            "process a specific bank statement",
        ),
        (
            "monopoly . --output <dir>",
            "saves all results to a specific directory",
        ),
        (
            "monopoly . --pprint",
            "prints results instead of saving them",
        ),
        (
            "monopoly --help",
            "show more options and other usage information",
        ),
    ]
    margin = max(len(cmd) for cmd, _ in commands)
    for cmd, desc in commands:
        styled_cmd = click.style(
            text=f"{cmd}{' ' * (margin - len(cmd))}",
            fg="white",
            bg="bright_black",
            bold=True,
        )
        click.echo(message=f"    {styled_cmd} {desc}")
    click.echo()



import logging
import re
from functools import cached_property

import fitz

from monopoly.config import DebitStatementConfig
from monopoly.constants import AccountType, StatementFields
from monopoly.pdf import PdfPage

from .base import BaseStatement, SafetyCheckError, Transaction

logger = logging.getLogger(__name__)


class DebitStatement(BaseStatement):
    """
    A dataclass representation of a debit statement
    """

    def __init__(
        self,
        document: fitz.Document,
        pages: list[PdfPage],
        debit_config: DebitStatementConfig,
    ):
        self.config = debit_config
        self.statement_type = AccountType.DEBIT
        super().__init__(document, pages, debit_config)

    def _process_line(
        self,
        line: str,
        lines: list[str],
        idx: int,
        pattern: re.Pattern,
    ) -> Transaction | None:
        if match := pattern.search(line):
            groupdict = match.groupdict()
            suffix = self.get_debit_suffix(line, pattern)

            # if the entry is a credit then add parenthesis around
            # the amount, which causes the Transaction class to
            # treat the amount as a credit / 'positive' transaction
            if suffix == "CR":
                line = line.replace(groupdict["amount"], f"({groupdict['amount']})")
                lines[idx] = line

            # continue with normal line processing
            return super()._process_line(line, lines, idx, pattern)
        return None

    def get_debit_suffix(self, line: str, pattern: re.Pattern):
        """
        Gets the accounting suffix for debit card statements

        This is necessary, since the amount in the row does not have any
        identifier apart from column position.
        """
        if match := pattern.search(line):
            amount = match[StatementFields.AMOUNT]
            pos = line.find(amount)
            withdrawal_diff = abs(pos - self.withdrawal_pos)
            deposit_diff = abs(pos - self.deposit_pos)
            if withdrawal_diff > deposit_diff:
                return "CR"
            return "DR"
        return None

    @cached_property
    def debit_header(self) -> str | None:
        if self.config and self.config.debit_statement_identifier:
            for line in self.pages[0].lines:
                if re.search(self.config.debit_statement_identifier, line):
                    return line.lower()
        return None

    @cached_property
    def withdrawal_pos(self) -> int:
        return self.get_header_pos("withdrawal")

    @cached_property
    def deposit_pos(self) -> int:
        return self.get_header_pos("deposit")

    def get_header_pos(self, column_name: str) -> int:
        """
        Returns position of the 'WITHDRAWAL' or 'DEPOSIT' header on a bank statement

        An assumption is made here that numbers are right aligned, meaning
        that if an amount matches with the end of the withdrawal string position,
        the item is in fact a withdrawal

        e.g.
        ```
        DATE         DESCRIPTION          WITHDRAWAL         DEPOSIT
        15 OCT       bill payment             322.07
        16 OCT       item                                     123.12
        ```
        """
        if self.debit_header:
            result = self.debit_header.find(column_name) + len(column_name)
        if not result:
            raise ValueError(
                f"Debit header {column_name} missing in {self.debit_header}"
            )
        return result

    def perform_safety_check(self) -> bool:
        """
        Checks that debit and credit transaction sums
        exist as a number within the statement
        """
        df = self.df
        amount = StatementFields.AMOUNT

        numbers = self.get_all_numbers_from_document()

        # add a zero, for cases where the debit statement
        # either is completely debit or credit transactions
        numbers.update([0])

        debit_sum = round(abs(df[df[amount] > 0][amount].sum()), 2)
        credit_sum = round(abs(df[df[amount] < 0][amount].sum()), 2)

        result = all([debit_sum in numbers, credit_sum in numbers])

        if not result:
            raise SafetyCheckError(self.failed_safety_message)

        return result



from monopoly.processors import ExampleBankProcessor


def example():
    """Example showing how monopoly can be used to extract data from
    a single bank statement
    """
    processor = ExampleBankProcessor(
        file_path="src/monopoly/examples/example_statement.pdf",
    )

    # This runs pdftotext on the PDF and
    # extracts transactions as raw text
    statement = processor.extract()

    # Dates are converted into an ISO 8601 date format
    transformed_df = processor.transform(statement)
    print(transformed_df)

    # Parsed transactions writen to a CSV file in the "example" directory
    processor.load(transformed_df, statement, output_directory="src/monopoly/examples")


if __name__ == "__main__":
    example()
